import abc
import hashlib
from ct.proto import certificate_pb2

def name_attribute(name):
    att = certificate_pb2.DNAttribute()
    att.type, att.value = "CN", name
    return att

SAMPLE_CERT = certificate_pb2.X509Description()
SAMPLE_CERT.der = "hello\x00"
SAMPLE_CERT.subject.extend([name_attribute("com.example")])
SAMPLE_CERT.sha256_hash = hashlib.sha256("hello\x00").digest()
FOUR_CERTS = []
for i in range(0, 4):
    tmp = certificate_pb2.X509Description()
    tmp.der = "hello-%d" % i
    tmp.subject.extend([name_attribute("com.domain-%d" % i)])
    tmp.sha256_hash = hashlib.sha256("hello-%d" % i).digest()
    FOUR_CERTS.append((tmp, i))
# This class provides common tests for all cert database implementations.
# It only inherits from object so that unittest won't attempt to run the test_*
# methods on this class. Derived classes should use multiple inheritance
# from CertDBTest and unittest.TestCase to get test automation.
class CertDBTest(object):
    """All cert database tests should derive from this class as well as
    unittest.TestCase."""
    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def db(self):
        """Derived classes must override to initialize a database."""
        pass

    def test_store_lookup_cert(self):
        self.db().store_cert_desc(SAMPLE_CERT, 0, 0)
        sha256_hash = hashlib.sha256("hello\x00").digest()
        cert = self.db().get_cert_by_sha256_hash(sha256_hash)
        self.assertEqual(SAMPLE_CERT.der, cert)

    def test_lookup_returns_none_if_no_match(self):
        self.db().store_cert_desc(SAMPLE_CERT, 0, 0)
        sha256_hash = hashlib.sha256("bye").digest()
        self.assertIsNone(self.db().get_cert_by_sha256_hash(sha256_hash))

    def test_store_cert_desc_ignores_duplicate(self):
        self.db().store_cert_desc(SAMPLE_CERT, 0, 0)
        sha256_hash = hashlib.sha256("hello\x00").digest()
        cert = self.db().get_cert_by_sha256_hash(sha256_hash)
        self.assertEqual("hello\x00", cert)

        # Store again
        hello2 = certificate_pb2.X509Description()
        hello2.der = SAMPLE_CERT.der
        hello2.subject.extend([name_attribute("com.example2")])
        hello2.sha256_hash = sha256_hash
        self.db().store_cert_desc(hello2, 1, 0)
        certs = [c for c in self.db().scan_certs_by_subject("example.com")]
        self.assertEqual(1, len(certs))
        self.assertEqual(SAMPLE_CERT.der, certs[0])
        certs = [c for c in self.db().scan_certs_by_subject("example2.com")]
        self.assertEqual(0, len(certs))

    def test_store_cert_descs_stores_all(self):
        self.db().store_certs_desc(FOUR_CERTS, 0)

        for i in range(4):
            sha256_hash = hashlib.sha256("hello-%d" % i).digest()
            cert = self.db().get_cert_by_sha256_hash(sha256_hash)
            self.assertEqual("hello-%d" % i, cert)

    def test_scan_certs_finds_all_certs(self):
        self.db().store_certs_desc(FOUR_CERTS, 0)

        certs = {c for c in self.db().scan_certs()}
        self.assertEqual(4, len(certs))
        for i in range(4):
            self.assertTrue("hello-%d" % i in certs)

    def test_scan_certs_honours_limit(self):
        self.db().store_certs_desc(FOUR_CERTS, 0)

        certs = {c for c in self.db().scan_certs(limit=2)}
        self.assertEqual(2, len(certs))

    def test_scan_certs_by_subject_finds_cert(self):
        self.db().store_certs_desc(FOUR_CERTS, 0)

        for i in range(4):
            matches = [c for c in self.db().scan_certs_by_subject(
                    "domain-%d.com" % i)]
            self.assertEqual(1, len(matches))
            self.assertEqual("hello-%d" % i, matches[0])

    def test_scan_certs_by_subject_honours_limit(self):
        for i in range(0, 4):
            cert = certificate_pb2.X509Description()
            cert.der = "hello-%d" % i
            cert.subject.extend([name_attribute("com.domain")])
            cert.sha256_hash = str(i)
            self.db().store_cert_desc(cert, i, 0)

        matches = {c for c in self.db().scan_certs_by_subject("domain.com",
                                                              limit=2)}
        self.assertEqual(2, len(matches))

    def test_scan_certs_by_subject_finds_by_common_name(self):
        for i in range(4):
            cert = certificate_pb2.X509Description()
            cert.der = "hello-%d" % i
            cert.subject.extend([name_attribute("Trusty CA %d" % i)])
            cert.sha256_hash = str(i)
            self.db().store_cert_desc(cert, i, 0)

        for i in range(4):
            matches = [c for c in self.db().scan_certs_by_subject(
                                                            "Trusty CA %d" % i)]
            self.assertEqual(1, len(matches))
            self.assertEqual("hello-%d" % i, matches[0])

    def test_scan_certs_by_subject_finds_by_all_names(self):
        cert = certificate_pb2.X509Description()
        cert.der = "hello"
        cert.subject.extend([name_attribute("com.%d" % i) for i in range(4)])
        cert.sha256_hash = 'asdf'
        self.db().store_cert_desc(cert, 0, 0)

        for i in range(4):
            matches = [c for c in self.db().scan_certs_by_subject("%d.com" % i)]
            self.assertEqual(1, len(matches))
            self.assertEqual("hello", matches[0])

    def test_scan_certs_by_subject_finds_by_prefix(self):
        for i in range(4):
            cert = certificate_pb2.X509Description()
            cert.der = "hello-%d" % i
            cert.subject.extend([name_attribute("com.example" + (".%d" % i)*i)])
            cert.sha256_hash = str(i)
            self.db().store_cert_desc(cert, i, 0)

        matches = {c for c in self.db().scan_certs_by_subject("example.com")}
        self.assertEqual(4, len(matches))
        for i in range(4):
            self.assertTrue("hello-%d" % i in matches)

    def test_scan_certs_by_subject_ignores_longer(self):
        cert = certificate_pb2.X509Description()
        cert.der = "hello"
        cert.subject.extend([name_attribute("example.com")])
        self.db().store_cert_desc(cert, 0, 0)

        matches = {c for c in self.db().scan_certs_by_subject(
                "mail.example.com")}
        self.assertEqual(0, len(matches))

    def test_scan_certs_by_subject_finds_by_literal_wildcard(self):
        cert = certificate_pb2.X509Description()
        cert.der = "hello-wild"
        cert.subject.extend([name_attribute("com.example.*")])
        cert.sha256_hash = 'cert'
        self.db().store_cert_desc(cert, 0, 0)
        cert2 = certificate_pb2.X509Description()
        cert2.der = "hello"
        cert2.subject.extend([name_attribute("com.example.www")])
        cert.sha256_hash = 'cert2'
        self.db().store_cert_desc(cert2, 1, 0)

        matches = [c for c in self.db().scan_certs_by_subject(
                "*.example.com")]
        self.assertEqual(1, len(matches))
        self.assertEqual("hello-wild", matches[0])

    def test_scan_certs_by_subject_ignores_wildcard_cert(self):
        for i in range(4):
            cert = certificate_pb2.X509Description()
            cert.der = "hello-%d" % i
            cert.subject.extend([name_attribute("com.example.%d" % i)])
            cert.sha256_hash = str(i)
            self.db().store_cert_desc(cert, i, 0)
        cert = certificate_pb2.X509Description()
        cert.der = "hello-wild"
        cert.subject.extend([name_attribute("com.example.*")])
        cert.sha256_hash = 'asdf'
        self.db().store_cert_desc(cert, 0, 0)

        matches = {c for c in self.db().scan_certs_by_subject(
                "2.example.com")}
        self.assertEqual(1, len(matches))
        self.assertTrue("hello-2" in matches)



"""Test meta-data and configuration."""

import os

CRYPTO_TEST_DATA_DIR = "ct/crypto/testdata/"
CERT_ANALYSIS_DATA_DIR = "ct/cert_analysis/test_data"

def get_test_file_path(filename):
    return os.path.join(os.curdir, CRYPTO_TEST_DATA_DIR, filename)

def get_tld_directory():
    return os.path.join(os.curdir, CERT_ANALYSIS_DATA_DIR)



import datetime
import base64
import hashlib
import math

def _write_cpp_header(f, include_guard):
    year = datetime.date.today().year
    f.write((
        "// Copyright %(year)d The Chromium Authors. All rights reserved.\n"
        "// Use of this source code is governed by a BSD-style license"
        " that can be\n"
        "// found in the LICENSE file.\n\n"
        "// This file is generated by print_log_list.py\n"
        "#ifndef %(include_guard)s\n"
        "#define %(include_guard)s\n\n"
        "#include <stddef.h>\n\n" %
        {"year": year, "include_guard": include_guard}))


def _write_log_info_struct_definition(f):
    f.write(
        "struct CTLogInfo {\n"
        "  const char* const log_key;\n"
        "  const size_t log_key_length;\n"
        "  const char* const log_name;\n"
        "  const char* const log_url;\n"
        "};\n\n")


def split_and_hexify_binary_data(bin_data):
    hex_data = "".join(["\\x%.2x" % ord(c) for c in bin_data])
    # line_width % 4 must be 0 to avoid splitting the hex-encoded data
    # across '\' which will escape the quotation marks.
    line_width = 68
    assert line_width % 4 == 0
    num_splits = int(math.ceil(len(hex_data) / float(line_width)))
    return ['"%s"' % hex_data[i * line_width:(i + 1) * line_width]
            for i in range(num_splits)]


def write_log_ids_array(f, log_ids, array_name):
    num_logs = len(log_ids)
    log_ids.sort()
    log_id_length = len(log_ids[0]) + 1
    f.write("// The list is sorted.\n")
    f.write("const char %s[][%d] = {\n" % (
            array_name, log_id_length))
    for i in range(num_logs):
        f.write("    ")
        split_hex_id = split_and_hexify_binary_data(log_ids[i])
        f.write("\n    ".join(split_hex_id))
        if (i < num_logs - 1):
            f.write(',\n')
    f.write('};\n\n')


def _write_cpp_footer(f, include_guard):
    f.write("\n#endif  // %s\n" % include_guard)


def _find_google_operator_id(json_log_list):
    goog_operator = filter(
        lambda op: op["name"] == "Google", json_log_list["operators"])
    assert(len(goog_operator) == 1)
    return goog_operator[0]["id"]


def generate_cpp_header(json_log_list, output_file):
    """Generate a header file of known logs to be included by Chromium."""
    include_guard = (output_file.upper().replace('.', '_').replace('/', '_')
                     + '_')
    f = open(output_file, "w")
    _write_cpp_header(f, include_guard)
    logs = json_log_list["logs"]
    list_code = []
    google_operator_id = _find_google_operator_id(json_log_list)
    google_log_ids = []
    for log in logs:
        log_key = base64.decodestring(log["key"])
        split_hex_key = split_and_hexify_binary_data(log_key)
        s = "    {"
        s += "\n     ".join(split_hex_key)
        s += ',\n     %d' % (len(log_key))
        s += ',\n     "%s"' % (log["description"])
        s += ',\n     "https://%s/"}' % (log["url"])
        list_code.append(s)

        # operated_by is a list, in practice we have not witnessed
        # a log co-operated by more than one operator. Ensure we take this
        # case into consideration if it ever happens.
        assert(len(log["operated_by"]) == 1)
        operator_id = log["operated_by"][0]
        if operator_id == google_operator_id:
            google_log_ids.append(hashlib.sha256(log_key).digest())

    _write_log_info_struct_definition(f)
    f.write("const CTLogInfo kCTLogList[] = {\n")
    f.write(",\n" . join(list_code))
    f.write("};\n")

    write_log_ids_array(f, google_log_ids, 'kGoogleLogIDs')

    _write_cpp_footer(f, include_guard)

